@@ -43,6 +43,10 @@
 ;;; o Rename "folder" to "mailbox".
 ;;; 
 
+(require 'rfc2104)
+(require 'base64)
+(require 'md5)
+
 (eval-when-compile (require 'cl))
 
 (eval-and-compile
@@ -50,15 +54,85 @@
   (unless (fboundp 'open-network-stream)
     (require 'tcp)))
 
+;;; New varibles (copyright jas):
+
+(defvar imap-data-capability nil
+  "Current server's capability list")
+
+(defvar imap-data-namespace nil
+  "Current server's namespace.")
+
+(defvar imap-cb-fetch-hook nil
+  "Hook called when receiving a FETCH response. Called with article NUM,
+FETCH and DATA response.")
+
+(defvar imap-default-user (user-login-name)
+  "Default username to use.")
+
+(defvar imap-streams '(kerberos4 ssl network)
+  "Priority of streams to consider when opening connection to
+server.")
+
+(defvar imap-stream-alist
+  '((kerberos4 imap-kerberos4s-p imap-kerberos4-open)
+    (ssl       imap-ssl-p        imap-ssl-open)
+    (network   imap-network-p    imap-network-open))
+  "Definition of network streams.
+
+(NAME CHECK OPEN)
+
+NAME names the stream, CHECK is a function returning non-nil if the
+server support the stream and OPEN is a function for opening the
+stream.")
+
+(defvar imap-authenticators '(kerberos4 cram-md5 login anonymous)
+  "Priority of authenticators to consider when authenticating to
+server.")
+
+(defvar imap-authenticator-alist 
+  '((kerberos4 imap-kerberos4a-p imap-kerberos4-auth)
+    (cram-md5  imap-cram-md5-p   imap-cram-md5-auth)
+    (login     imap-login-p      imap-login-auth)
+    (anonymous imap-anonymous-p  imap-anonymous-auth))
+  "Definition of authenticators.
+
+(NAME CHECK AUTHENTICATE)
+
+NAME names the authenticator. CHECK is a function returning non-nil if
+the server support the authenticator and AUTHENTICATE is a function
+for doing the actuall authentification.")
+
+;; Internal constants. Change theese and die.
+
+(defconst imap-default-port 143)
+(defconst imap-default-ssl-port 993)
+(defconst imap-default-stream 'network)
+(defconst imap-coding-system-for-read 'binary)
+(defconst imap-coding-system-for-write 'binary)
+(defconst imap-local-variables '(imap-server
+				 imap-port
+				 imap-auth
+				 imap-stream
+				 imap-username
+				 imap-password))
+
+;; Internal variables:
+
+(defvar imap-buffer-list nil
+  "List of buffers the imap library has created, which are destroyed
+on call to `imap-done'.")
+
+(defvar imap-stream nil)
+(defvar imap-auth nil)
+(defvar imap-server nil)
+(defvar imap-port nil)
+(defvar imap-username nil)
+(defvar imap-password nil)
+
+;;; Old varibles (see changelog for copyright status):
+
 ;;; External variables
 
-(defvar imap-default-port 143
-  "*Default port number to be used for IMAP connections.  This should
-probably be \"imap\", but a lot of machines lack the services entry.
-
-This can be overrided by the server definition imap-port, and is the
-prefered way of specifying this.")
-
 (defvar imap-convenient-group-prime 2999
   "*A convenient prime which will be used to set the size of the group hash.
 We have a lot of groups at CMU, so this should probably be adjusted down.")
@@ -67,61 +141,20 @@ We have a lot of groups at CMU, so this should probably be adjusted down.")
   "*A convenient prime which will be used to set the size of the folder
 (message) hash.")
 
-(defvar imap-open-stream nil
-  "*The name of a function to use for opening an imap stream. Defaults on
-nil to open a networked stream to the server.
-
-Examples; imap-open-imtest-stream, imap-open-ssl-stream.
-
-This can be overrided by the server definition imap-open-stram, and
-this is the prefered way of specifying this.")
-
-(defvar imap-auth-method nil
-  "*The name of a function to use for loging on to the server. Defaults on
-nil to plain text logins using the LOGIN command.
-
-Examples; imap-authenticate-cram-md5.
-
-This can be overried by the server definition imap-auth-method, and
-this is the prefered way of specifying this.")
-
 (defvar imap-eol "\r\n"
   "*The string sent to end a command.")
 
-;; remove?
-(defvar imap-default-name nil
-  "*Your name, should you choose to accept it.")
-
 (defvar imap-last-status nil
   "*Status returned by last IMAP command")
 
 (defvar imap-timeout 60
   "*Timeout in seconds to wait for server response.")
 
-(defvar imap-username nil
-  "Username for server. ")
-
-(defvar imap-password nil
-  "Password for server.")
-
-(defvar imap-cb-fetch-hook nil
-  "Hook called when receiving a FETCH response. Called with article NUM,
-FETCH and DATA response.")
-
 ;;; Internal variables
 
-(defvar imap-authinfo nil
-  "Buffer local variable which contains (user . password) for server.")
-
 (defvar imap-process nil
   "The active process for the current IMAP buffer.")
 
-(defvar imap-data-capability nil
-  "Current server's capability list")
-
-(defvar imap-data-namespace nil
-  "Current server's namespace.")
-
 (defvar imap-data-folder nil
   "Obarray which contains group information.")
 
@@ -149,12 +182,6 @@ FETCH and DATA response.")
 (defvar imap-connection-number 0
   "Unique tag char per connection.")
 
-(defvar imap-do-login t
-  "Wheter imap-authenticate should try to log in or not.
-
-This is normally only turned off by a `imap-open-stream' that does
-it's own authentication.")
-
 (defvar imap-cb-function-alist '((OK . imap-cb-response)
 				 (NO . imap-cb-response)
 				 (BAD . imap-cb-response)
@@ -186,12 +213,8 @@ it's own authentication.")
                       imap-data-capability
 		      imap-data-namespace
                       imap-data-folder
-		      imap-open-stream
-		      imap-auth-method
-                      imap-do-login
                       imap-message-data
                       imap-default-name
-		      imap-authinfo
                       imap-last-status
                       imap-process
                       imap-tag-num
@@ -207,34 +230,474 @@ any of these to `imap-open-server'.")
 (defvar imap-last  nil);"*imap-last*")  ; last line we attempted to parse
 (defvar imap-debug nil);"*imap-debug*") ; random debug spew
 
+;;; New functions (copyright jas):
+
+
+;; Server functions; stream stuff:
+
+(defun imap-kerberos4s-p (buffer)
+  (imap-capability 'AUTH=KERBEROS_V4 buffer))
+
+(defun imap-kerberos4-open (name buffer server port)
+  (setq port (or port imap-default-port))
+  (let* ((coding-system-for-read imap-coding-system-for-read)
+	 (coding-system-for-write imap-coding-system-for-write)
+	 (process (start-process name buffer "imtest" "-kp" server
+				 (number-to-string port))))
+    (with-current-buffer (process-buffer process)
+      (setq imap-eol "\n") ;; xxx
+      (when process
+	(message "Opening Kerberized IMAP connection...")
+	;; Result of authentication is a string: __Full privacy protection__
+	(while (and (memq (process-status process) '(open run))
+		    (goto-char (point-max))
+		    (not (or
+			  ;; if everything is ok, this will match...
+			  (re-search-backward "__\\(.*\\)__\n" nil t)
+			  ;; ...errors will match this
+			  (re-search-backward 
+			   "\\. \\([^OA][^KU][^T][^H].*\\)\r\n" nil t))))
+	  (accept-process-output process 1)
+	  (sit-for 1))
+	(let ((response (match-string 1)))
+	  (message "Kerberized IMAP connection: %s" response)
+	  (if (string-match "failed\\|NO\\|BAD" response)
+	      (progn
+		(imap-send-command-wait "LOGOUT")
+		(delete-process process)
+		response)
+	    (erase-buffer)
+	    process))))))
+  
+(defun imap-ssl-p (buffer)
+  nil)
+
+(defun imap-ssl-open-1 (name buffer server port extra-arg)
+  (setq port (or port imap-default-ssl-port))
+  (let* ((coding-system-for-read imap-coding-system-for-read)
+	 (coding-system-for-write imap-coding-system-for-write)
+	 (ssl-program-arguments (list extra-arg
+				      "-connect" 
+				      (format "%s:%d" imap-server port)))
+	 (proc (open-ssl-stream name buffer server port)))
+    (with-current-buffer buffer
+      (goto-char (point-min))
+      (while (and (memq (process-status proc) '(open run))
+		  (goto-char (point-min))
+		  (not (re-search-forward "^\r*\\* OK" nil t)))
+	(accept-process-output proc 1)
+	(goto-char (point-min)))
+      (beginning-of-line)
+      (delete-region (point-min) (point))
+      (when (memq (process-status proc) '(open run))
+	proc))))
+
+(defun imap-ssl-open (name buffer server port)
+  (message "Opening SSL3 connection...")
+  (let ((ret (imap-ssl-open-1 name buffer server port "-ssl3")))
+    (if ret
+	ret
+      (message "Opening SSL2 connection...")
+      (imap-ssl-open-1 name buffer server port "-ssl2"))))
+
+(defun imap-network-p (buffer)
+  t)
+
+(defun imap-network-open (name buffer server port)
+  (setq port (or port imap-default-port))
+  (let ((coding-system-for-read imap-coding-system-for-read)
+	(coding-system-for-write imap-coding-system-for-write))
+    (condition-case ()
+	(open-network-stream name buffer server port)
+      (error nil)
+      (quit nil))))
+  
+;; Server functions; authenticator stuff:
+
+(defun imap-interactive-login (buffer loginfunc)
+  "Login to server in BUFFER. LOGINFUNC is passed a username and a
+password, it should return t if it where sucessful authenticating
+itself to the server, nil otherwise. Returns t if login was
+successful, nil otherwise."
+  (with-current-buffer buffer
+    (make-variable-buffer-local 'imap-username)
+    (make-variable-buffer-local 'imap-password)
+    (let (user passwd ret)
+      (condition-case ()
+	  (while (or (not user) (not passwd))
+	    (setq user (or imap-username
+			   (read-from-minibuffer 
+			    (concat "IMAP username for " imap-server ": ")
+			    (or user imap-default-user))))
+	    (setq passwd (or imap-password
+			     (imap-read-passwd
+			      (concat "IMAP password for " user "@" 
+				      imap-server ": "))))
+	    (when (and user passwd)
+	      (if (funcall loginfunc user passwd)
+		  (progn
+		    (setq ret t
+			  imap-username user)
+		    (if (and (not imap-password)
+			     (y-or-n-p "Store password for this session? "))
+			(setq imap-password passwd)))
+		(message "Login failed...")
+		(setq passwd nil)
+		(sit-for 1))))
+	(quit (with-current-buffer buffer
+		(setq user nil
+		      passwd nil))))
+      ret)))
+
+(defun imap-kerberos4a-p (buffer)
+  (imap-capability 'AUTH=KERBEROS_V4 buffer))
+
+(defun imap-kerberos4-auth (buffer)
+  (eq imap-stream 'kerberos4))
+
+(defun imap-cram-md5-p (buffer)
+  (imap-capability 'AUTH=CRAM-MD5 buffer))
+
+(defun imap-cram-md5-auth (buffer)
+  "Login to server using the AUTH CRAM-MD5 method."
+  (imap-interactive-login
+   buffer
+   (lambda (user passwd)
+     (imap-ok-p
+      (imap-send-command-wait
+       (list
+	"AUTHENTICATE CRAM-MD5"
+	(lambda (challenge)
+	  (let* ((decoded (base64-decode challenge))
+		 (hash (rfc2104-hash 'md5 64 16 passwd decoded))
+		 (response (concat user " " hash))
+		 (encoded (base64-encode response)))
+	    encoded))))))))
+
+(defun imap-login-p (buffer)
+  (not (imap-capability 'X-LOGIN-CMD-DISABLED buffer)))
+
+(defun imap-login-auth (buffer)
+  "Login to server using the LOGIN command."
+  (imap-interactive-login buffer 
+			  (lambda (user passwd)
+			    (imap-ok-p (imap-send-command-wait 
+					(concat "LOGIN \"" user "\" \"" 
+						passwd "\""))))))
+
+(defun imap-anonymous-p (buffer)
+  t)
+
+(defun imap-anonymous-open (buffer)
+  (with-current-buffer buffer
+    (imap-ok-p (imap-send-command-wait
+		(concat "LOGIN anonymous \"" (concat (user-login-name) "@" 
+						     (system-name)) "\"")))))
+
+;; Server functions:
+
+(defun imap-open-1 (buffer)
+  (with-current-buffer buffer
+    (setq imap-process (funcall (nth 2 (assq imap-stream imap-stream-alist))
+				"imap" buffer imap-server imap-port))
+    (when imap-process
+      (set-marker (process-mark imap-process) (point-min))
+      (set-process-filter imap-process 'imap-arrival-filter)
+      (setq imap-current-folder nil
+	    imap-current-message nil)
+      ;; legacy:
+      (setq imap-tag-char (int-char (+ (char-int ?A) 
+				       (% imap-connection-number 26))))
+      (setq imap-connection-number (1+ imap-connection-number))
+      (setq imap-data-folder (make-vector imap-convenient-group-prime 0))
+      ;; end legacy
+      imap-process)))
+
 (defsubst imap-disable-multibyte ()
   "Enable multibyte in the current buffer."
   (when (fboundp 'set-buffer-multibyte)
     (set-buffer-multibyte nil)))
 
+(defun imap-open (server &optional port stream auth buffer)
+  (setq buffer (or buffer (format " *imap* %s:%d" server port)))
+  (unless (get-buffer buffer)
+    (setq imap-buffer-list (cons buffer imap-buffer-list)))
+  (with-current-buffer (get-buffer-create buffer)
+    (let (stream-changed)
+      (if (imap-opened buffer)
+	  (imap-close buffer)
+	(mapc 'make-variable-buffer-local imap-locals)
+	(mapc 'make-variable-buffer-local imap-local-variables)
+	(buffer-disable-undo)
+	(imap-disable-multibyte)) ;; we shouldn't do this
+      (if server (setq imap-server server))
+      (if port (setq imap-port port))
+      (if auth (setq imap-auth auth))
+      (if stream (setq imap-stream stream))
+      (if (and imap-stream imap-auth)
+	  (imap-open-1 buffer)
+	;; Determine network stream and/or auth mechanism to use
+	(if (not (let ((imap-stream (or imap-stream imap-default-stream)))
+		   (imap-open-1 buffer)))
+	    ;; clean up and exit
+	    (when (memq buffer imap-buffer-list)
+	      (kill-buffer buffer)
+	      (setq imap-buffer-list (delq buffer imap-buffer-list)))
+	  ;; Choose stream.
+	  (when (null imap-stream)
+	    (let ((streams imap-streams))
+	      (while (setq stream (pop streams))
+		(if (funcall (nth 1 (assq stream imap-stream-alist)) buffer)
+		    (setq stream-changed (not (eq
+					       imap-stream
+					       (or stream 
+						   imap-default-stream)))
+			  imap-stream stream
+			  streams nil)))
+	      (unless imap-stream
+		(error "Couldn't figure out a stream for server"))))
+	  ;; Choose authenticator
+	  (when (null imap-auth)
+	    (let ((auths imap-authenticators))
+	      (while (setq auth (pop auths))
+		(if (funcall (nth 1 (assq auth imap-authenticator-alist)) 
+			     buffer)
+		    (setq imap-auth auth
+			  auths nil)))
+	      (unless imap-auth
+		(error "Couldn't figure out authenticator for server"))))
+	  (when stream-changed
+	    (imap-close buffer)
+	    (imap-open-1 buffer))))))
+  (if (imap-opened buffer)
+      buffer))
+
+(defun imap-opened (&optional buffer)
+  (and (setq buffer (get-buffer (or buffer (current-buffer))))
+       (buffer-live-p buffer)
+       (with-current-buffer buffer
+	 (and imap-process
+	      (memq (process-status imap-process) '(open run))))))
+
+(defun imap-authenticate (buffer &optional user passwd)
+  (with-current-buffer buffer
+    (make-variable-buffer-local 'imap-username)
+    (make-variable-buffer-local 'imap-password)
+    (if user (setq imap-username user))
+    (if passwd (setq imap-password passwd))
+    (funcall (nth 2 (assq imap-auth imap-authenticator-alist)) buffer)))
+  
+(defun imap-close (&optional buffer)
+  (with-current-buffer (or buffer (current-buffer))
+    (if (not (and (imap-opened)
+		  (imap-send-command-wait "LOGOUT")))
+	(error "Server %s didn't let me log out" imap-server)
+      (when (memq (process-status imap-process) '(open run))
+	(delete-process imap-process))
+      (setq imap-current-folder nil
+	    imap-current-message nil
+	    imap-process nil)
+      (erase-buffer)
+      t)))
+
+(defun imap-capability (&optional identifier buffer)
+  "Return a list of identifiers which the server support. If IDENTIFIER,
+return non-nil if it's among the servers capabilities."
+  (with-current-buffer (or buffer (current-buffer))
+    (unless imap-data-capability
+      (unless (imap-ok-p (imap-send-command-wait "CAPABILITY"))
+	(setq imap-data-capability '(IMAP2))))
+    (if identifier
+	(memq identifier imap-data-capability)
+      imap-data-capability)))
+
+(defun imap-namespace-get (&optional buffer)
+  "Return server's namespace."
+  (with-current-buffer (or buffer (current-buffer))
+    (unless imap-data-namespace
+      (when (imap-capability 'NAMESPACE)
+	(imap-send-command-wait "NAMESPACE")))
+    imap-data-namespace))
+
+
+;; Mailbox functions:
+
+(defun imap-mailbox-map (func &optional buffer)
+  "Call (func FOLDER) for each folder in `imap-data-folder', returning
+a sequence."
+  (with-current-buffer (or buffer (current-buffer))
+    (let (result)
+      (mapatoms 
+       (lambda (s)
+	 (push (funcall func (symbol-name s)) result))
+       imap-data-folder)
+      result)))
+
+(defun imap-mailbox-reset (&optional buffer)
+  (with-current-buffer (or buffer (current-buffer)) 
+    (setq imap-data-folder (make-vector imap-convenient-group-prime 0))))
+
+(defun imap-mailbox-select (folder &optional buffer examine)
+  (with-current-buffer (or buffer (current-buffer))
+    ;; xxx: check SELECT/EXAMINE status!  This is BAD.
+    (unless (string= folder imap-current-folder)   
+      (setq imap-current-folder folder)
+      (if (imap-ok-p (imap-send-command-wait
+		      (concat (if examine "EXAMINE" "SELECT") " " folder)))
+	  (setq imap-message-data (make-vector imap-convenient-folder-prime 0))
+	;; Failed SELECT unselects the current group
+	(setq imap-current-folder nil
+	      imap-message-data nil)))
+    imap-current-folder))
+
+(defun imap-mailbox-unselect (&optional group buffer)
+  "Close current folder in BUFFER, without expunging articles."
+  (with-current-buffer (or buffer (current-buffer))
+    (when (or (and (imap-capability 'UNSELECT)
+		   (imap-ok-p (imap-send-command-wait "UNSELECT")))
+	      (and (imap-ok-p 
+		    (imap-send-command-wait (concat "EXAMINE " 
+						    (or group
+							imap-current-folder))))
+		   (imap-ok-p (imap-send-command-wait "CLOSE"))))
+      (setq imap-current-folder nil
+	    imap-message-data nil)
+      t)))
+
+(defun imap-mailbox-close (&optional buffer)
+  "Expunge articles and close current folder in BUFFER."
+  (with-current-buffer (or buffer (current-buffer))
+    (when (imap-ok-p (imap-send-command-wait "CLOSE"))
+      (setq imap-current-folder nil
+	    imap-message-data nil)
+      t)))
+
+(defun imap-mailbox-lsub (&optional reference buffer)
+  "Return a list of strings of subscribed mailboxes on server in
+BUFFER. REFERENCE is the implementation-specific string that has to be
+passed to LSUB."
+  (with-current-buffer (or buffer (current-buffer))
+    (imap-mailbox-reset)
+    (when (imap-ok-p (imap-send-command-wait 
+		      (concat "LSUB \"" reference "\" \"*\"")))
+      (imap-mailbox-map 'identity))))
+
+(defun imap-mailbox-list (&optional root have-delimiter reference buffer)
+  "List all mailboxes that starts with ROOT in BUFFER. If
+HAVE-DELIMITER is non-nil, a hierarchy delimiter is not added to
+ROOT. REFERENCE is the implementation-specific string that has to be
+passed to LIST."
+  (with-current-buffer (or buffer (current-buffer))
+    (imap-mailbox-reset)
+    ;; Find hierarchy separator
+    (unless have-delimiter
+      (imap-send-command-wait (concat "LIST \"" reference "\" \"" root "\"")))
+    (when (imap-ok-p 
+	   (imap-send-command-wait 
+	    (concat "LIST \"" reference "\" \"" root
+		    (when (and (not have-delimiter) root)
+		      (imap-folder-get 'delimiter root))
+		    "%\"")))
+      (imap-mailbox-map 'identity))))
+
+(defun imap-mailbox-subscribe (mailbox &optional buffer)
+  "Send the SUBSCRIBE command on the mailbox to server in
+BUFFER. Returns non-nil if successful."
+  (with-current-buffer (or buffer (current-buffer))
+    (imap-ok-p (imap-send-command-wait (concat "SUBSCRIBE " mailbox)))))
+
+(defun imap-mailbox-unsubscribe (mailbox &optional buffer)
+  "Send the SUBSCRIBE command on the mailbox to server in
+BUFFER. Returns non-nil if successful."
+  (with-current-buffer (or buffer (current-buffer))
+    (imap-ok-p (imap-send-command-wait (concat "UNSUBSCRIBE " mailbox)))))
+
+
+;; Message functions:
+
+(defun imap-search (predicate &optional buffer)
+  (with-current-buffer (or buffer (current-buffer))
+    (imap-folder-set 'search nil)
+    (when (imap-ok-p (imap-send-command-wait (concat "UID SEARCH " predicate)))
+      (imap-folder-get 'search))))
+
+(defun imap-message-flags-set (articles flags &optional buffer silent)
+  (when (and articles flags)
+    (with-current-buffer (or buffer (current-buffer))
+      (imap-ok-p (imap-send-command-wait
+		  (concat "UID STORE " articles
+			  " FLAGS" (if silent ".SILENT") " (" flags ")"))))))
+
+(defun imap-message-flags-del (articles flags &optional buffer silent)
+  (when (and articles flags)
+    (with-current-buffer (or buffer (current-buffer))
+      (imap-ok-p (imap-send-command-wait
+		  (concat "UID STORE " articles
+			  " -FLAGS" (if silent ".SILENT") " (" flags ")"))))))
+
+(defun imap-message-flags-add (articles flags &optional buffer silent)
+  (when (and articles flags)
+    (with-current-buffer (or buffer (current-buffer))
+      (imap-ok-p (imap-send-command-wait
+		  (concat "UID STORE " articles
+			  " +FLAGS" (if silent ".SILENT") " (" flags ")"))))))
+
+(defun imap-message-reset (&optional buffer)
+  (with-current-buffer (or buffer (current-buffer)) 
+    (setq imap-message-data (make-vector imap-convenient-folder-prime 0))))
+
+;; Stuff:
+
+(defun imap-cb-capability (code &rest capabilities)
+  (setq imap-data-capability capabilities))
+
+(defun imap-cb-acl (code group &rest acls)
+  (imap-folder-set 'acl acls (symbol-name group)))
+
+(defun imap-cb-namespace (code &rest namespace)
+  (setq imap-data-namespace namespace))
+
+;;; Old functions (see changelog for copyright status):
+
 (when imap-debug ; (untrace-all)
   (require 'trace)
   (buffer-disable-undo (get-buffer-create imap-debug))
   (imap-disable-multibyte)
   (mapc (lambda (f) (trace-function-background f imap-debug)) 
-        '(imap-open-server
-          imap-close-server
-          imap-server-opened
-	  imap-select-folder
-	  imap-unselect-folder
-	  imap-expunge-close-folder
+        '(imap-kerberos4s-p
+	  imap-kerberos4-open
+	  imap-ssl-p
+	  imap-ssl-open-1
+	  imap-ssl-open
+	  imap-network-p
+	  imap-network-open
+	  imap-interactive-login
+	  imap-kerberos4a-p
+	  imap-kerberos4-auth
+	  imap-cram-md5-p
+	  imap-cram-md5-auth
+	  imap-login-p
+	  imap-login-auth
+	  imap-anonymous-p
+	  imap-anonymous-open
+	  imap-open-1
+	  imap-open
+	  imap-opened
+          imap-close
+	  imap-mailbox-select
+	  imap-mailbox-unselect
+	  imap-mailbox-expunge-close
           imap-send-command
           imap-send-command-wait
           imap-send-commands-wait
 ;	  imap-ok-p
 	  imap-wait-for-tag
-	  imap-capability-get
+	  imap-capability
 	  imap-namespace-get
-	  imap-authinfo-get
           imap-folder-set
           imap-folder-get
           imap-folder-plist
-	  imap-folder-reset
+	  imap-mailbox-reset
           imap-dispatch
 	  imap-authenticate
 	  imap-authenticate-login
@@ -325,86 +788,10 @@ See also `plist-to-alist'."
 
 ;;; Interface functions
 
-(defun imap-server-opened (&optional buffer)
-  (with-current-buffer (or buffer (current-buffer))
-    (and imap-process
-         (member (process-status imap-process) '(open run)))))
-
-(defun imap-close-server (&optional buffer autologout)
-  "Logout if needed and close down the process.  Clean out buffer.
-Ensure all `imap-locals' are local and reset them to their default
-values such that the buffer will be suitable for opening a new server."
-  ;; What is this for???
-  (setq buffer (get-buffer (or buffer (current-buffer))))
-  (when buffer
-    (with-current-buffer buffer
-      (mapc 'make-variable-buffer-local imap-locals) ; just in case
-      (when imap-process
-        (and (member (process-status imap-process) '(open run))
-             (imap-send-command-wait "LOGOUT"))
-        (delete-process imap-process))
-      (mapc (lambda (local) (set local (default-value local))) imap-locals)
-      (erase-buffer)
-      t)))
-
 (defun imap-current-server (&optional buffer)
   (with-current-buffer (or buffer (current-buffer)) 
     imap-current-server))
 
-(defun imap-authenticate-login (server &optional buffer)
-  "Login to server using the LOGIN command."
-  (with-current-buffer (or buffer (current-buffer))
-    (and (imap-authinfo-get server)
-	 (imap-ok-p (imap-send-command-wait 
-		       (concat "LOGIN \"" (car imap-authinfo) 
-			       "\" \"" (cdr imap-authinfo) "\""))))))
-
-(defun imap-authenticate-cram-md5 (server &optional buffer)
-  "Login to server using the AUTH CRAM-MD5 method."
-  (require 'rfc2104)
-  (require 'md5)
-  (with-current-buffer (or buffer (current-buffer))
-    (and (imap-authinfo-get server)
-	 (memq 'AUTH=CRAM-MD5 (imap-capability-get))
-	 (imap-ok-p 
-	  (imap-send-command-wait
-	   (list 
-	    "AUTHENTICATE CRAM-MD5"
-	    (lambda (challenge)
-	      (let* ((decoded (base64-decode challenge))
-		     (hash (rfc2104-hash 'md5 64 16 (cdr imap-authinfo) decoded))
-		     (response (concat (car imap-authinfo) " " hash))
-		     (encoded (base64-encode response)))
-		encoded))))))))
-
-(defun imap-authenticate (server &optional buffer)
-  (if (not imap-do-login)
-      t
-    (with-current-buffer (or buffer (current-buffer))
-      (if imap-auth-method
-	  (funcall imap-auth-method server buffer)
-	(imap-authenticate-login server buffer)))))
-
-(defun imap-open-server (server &optional port buffer local-defs)
-  (with-current-buffer (get-buffer-create (or buffer (current-buffer)))
-    (buffer-disable-undo)
-    (imap-disable-multibyte)
-    (imap-close-server) ; makes vars local, sets them to their defaults, erases
-    (mapc (lambda (ld) (set (car ld) (cdr ld))) local-defs)
-    (when (setq imap-process 
-                (imap-open-stream "imap" (current-buffer) 
-                                  server (or port imap-default-port)))
-      (set-marker (process-mark imap-process) (point-min))
-      (set-process-filter imap-process 'imap-arrival-filter)
-      (setq imap-current-server server)
-      ;; Give each connection a more or less unique letter just so the log
-      ;; is easy to read
-      (setq imap-tag-char (int-char (+ (char-int ?A) 
-				       (% imap-connection-number 26))))
-      (setq imap-connection-number (1+ imap-connection-number))
-      (setq imap-data-folder (make-vector imap-convenient-group-prime 0))
-      (current-buffer))))
-
 ;; If there is a need for sending commands without a callback, then
 ;; have `imap-send-command-wait'ing commands pass 
 ;; `imap-cb-tag-default' itself.  Maybe `imap-wait-for-tag' should
@@ -504,126 +891,8 @@ in order.  See `imap-send-command-wait'."
     (setq imap-last-status nil)
     t))
 
-(defun imap-search (predicate &optional buffer)
-  (with-current-buffer (or buffer (current-buffer))
-    (imap-folder-set 'search nil)
-    (when (imap-ok-p (imap-send-command-wait (concat "UID SEARCH " predicate)))
-      (imap-folder-get 'search))))
-
-(defun imap-store-flags-set (articles flags &optional buffer silent)
-  (when (and articles flags)
-    (with-current-buffer (or buffer (current-buffer))
-      (imap-ok-p (imap-send-command-wait
-		  (concat "UID STORE " articles
-			  " FLAGS" (if silent ".SILENT") " (" flags ")"))))))
-
-(defun imap-store-flags-del (articles flags &optional buffer silent)
-  (when (and articles flags)
-    (with-current-buffer (or buffer (current-buffer))
-      (imap-ok-p (imap-send-command-wait
-		  (concat "UID STORE " articles
-			  " -FLAGS" (if silent ".SILENT") " (" flags ")"))))))
-
-(defun imap-store-flags-add (articles flags &optional buffer silent)
-  (when (and articles flags)
-    (with-current-buffer (or buffer (current-buffer))
-      (imap-ok-p (imap-send-command-wait
-		  (concat "UID STORE " articles
-			  " +FLAGS" (if silent ".SILENT") " (" flags ")"))))))
-
-(defun imap-select-folder (folder &optional buffer examine)
-  (with-current-buffer (or buffer (current-buffer))
-    ;; xxx: check SELECT/EXAMINE status!  This is BAD.
-    (unless (string= folder imap-current-folder)   
-      (setq imap-current-folder folder)
-      (if (imap-ok-p (imap-send-command-wait
-		      (concat (if examine "EXAMINE" "SELECT") " " folder)))
-	  (setq imap-message-data (make-vector imap-convenient-folder-prime 0))
-	;; Failed SELECT unselects the current group
-	(setq imap-current-folder nil
-	      imap-message-data nil)))
-    imap-current-folder))
-
-(defun imap-unselect-folder (&optional group buffer)
-  "Close current folder in BUFFER, without expunging articles."
-  (with-current-buffer (or buffer (current-buffer))
-    (when (or (and (memq 'UNSELECT (imap-capability-get))
-		   (imap-ok-p (imap-send-command-wait "UNSELECT")))
-	      (and (imap-ok-p 
-		    (imap-send-command-wait (concat "EXAMINE " 
-						    (or group
-							imap-current-folder))))
-		   (imap-ok-p (imap-send-command-wait "CLOSE"))))
-      (setq imap-current-folder nil
-	    imap-message-data nil)
-      t)))
-
-(defun imap-expunge-close-folder (&optional buffer)
-  "Expunge articles and close current folder in BUFFER."
-  (with-current-buffer (or buffer (current-buffer))
-    (when (imap-ok-p (imap-send-command-wait "CLOSE"))
-      (setq imap-current-folder nil
-	    imap-message-data nil)
-      t)))
-
-(defun imap-folder-lsub (&optional reference buffer)
-  "Return a list of strings of subscribed mailboxes on server in
-BUFFER. REFERENCE is the implementation-specific string that has to be
-passed to LSUB."
-  (with-current-buffer (or buffer (current-buffer))
-    (imap-folder-reset)
-    (when (imap-ok-p (imap-send-command-wait 
-		      (concat "LSUB \"" reference "\" \"*\"")))
-      (imap-folder-map 'identity))))
-
-(defun imap-folder-list (&optional root have-delimiter reference buffer)
-  "List all mailboxes that starts with ROOT in BUFFER. If
-HAVE-DELIMITER is non-nil, a hierarchy delimiter is not added to
-ROOT. REFERENCE is the implementation-specific string that has to be
-passed to LIST."
-  (with-current-buffer (or buffer (current-buffer))
-    (imap-folder-reset)
-    ;; Find hierarchy separator
-    (unless have-delimiter
-      (imap-send-command-wait (concat "LIST \"" reference "\" \"" root "\"")))
-    (when (imap-ok-p 
-	   (imap-send-command-wait 
-	    (concat "LIST \"" reference "\" \"" root
-		    (when (and (not have-delimiter) root)
-		      (imap-folder-get 'delimiter root))
-		    "%\"")))
-      (imap-folder-map 'identity))))
-
-(defun imap-folder-subscribe (mailbox &optional buffer)
-  "Send the SUBSCRIBE command on the mailbox to server in
-BUFFER. Returns non-nil if successful."
-  (with-current-buffer (or buffer (current-buffer))
-    (imap-ok-p (imap-send-command-wait (concat "SUBSCRIBE " mailbox)))))
-
-(defun imap-folder-unsubscribe (mailbox &optional buffer)
-  "Send the SUBSCRIBE command on the mailbox to server in
-BUFFER. Returns non-nil if successful."
-  (with-current-buffer (or buffer (current-buffer))
-    (imap-ok-p (imap-send-command-wait (concat "UNSUBSCRIBE " mailbox)))))
-
 ;;; Variable setters and getters
 
-(defun imap-capability-get (&optional buffer)
-  "Return a list of identifiers which the server support."
-  (with-current-buffer (or buffer (current-buffer))
-    (unless imap-data-capability
-      (unless (imap-send-command-wait "CAPABILITY")
-	(setq imap-data-capability '(IMAP2))))
-    imap-data-capability))
-
-(defun imap-namespace-get (&optional buffer)
-  "Return server's namespace."
-  (with-current-buffer (or buffer (current-buffer))
-    (unless imap-data-namespace
-      (when (memq 'NAMESPACE (imap-capability-get))
-	(imap-send-command-wait "NAMESPACE")))
-    imap-data-namespace))
-
 (defun imap-folder-plist (&optional folder buffer)
   "Set PROP to VALUE for FOLDER in BUFFER."
   (with-current-buffer (or buffer (current-buffer))
@@ -644,21 +913,6 @@ BUFFER. Returns non-nil if successful."
     (get (intern (or folder
                      imap-current-folder) imap-data-folder) prop)))
 
-(defun imap-folder-map (func &optional buffer)
-  "Call (func FOLDER) for each folder in `imap-data-folder', returning
-a sequence."
-  (with-current-buffer (or buffer (current-buffer))
-    (let (result)
-      (mapatoms 
-       (lambda (s)
-	 (push (funcall func (symbol-name s)) result))
-       imap-data-folder)
-      result)))
-
-(defun imap-folder-reset (&optional buffer)
-  (with-current-buffer (or buffer (current-buffer)) 
-    (setq imap-data-folder (make-vector imap-convenient-group-prime 0))))
-
 ;;; Internal functions
 
 (defun imap-read-passwd (prompt &rest args)
@@ -673,84 +927,6 @@ If ARGS, PROMPT is used as an argument to `format'."
 		 (autoload 'ange-ftp-read-passwd "ange-ftp"))
 	       'ange-ftp-read-passwd) prompt)))
 
-(defun imap-authinfo-get (server &optional buffer)
-  "Get user authentication information. Uses imap-username and/or
-imap-password. Asks the user if necessery. If successful, sets 
-imap-authinfo to (username . password)."
-  (with-current-buffer (or buffer (current-buffer))
-    (let (user passwd)
-      (setq user (or imap-username
-		     (read-from-minibuffer (concat "IMAP Name for " server 
-						   ": ")
-					   imap-default-name)))
-      (setq passwd (or imap-password
-		       (imap-read-passwd (concat "IMAP Password for " user "@"
-						 server ": "))))
-      (if (and user passwd)
-	  (progn
-	    (setq imap-authinfo (cons user passwd))
-	    t)
-	(setq imap-authinfo nil)))))
-
-(defun imap-open-stream (name buffer host &optional port)
-  (let ((coding-system-for-read 'binary)
-	(coding-system-for-write 'binary))
-    (if imap-open-stream
-	(funcall imap-open-stream name buffer host port)
-      (imap-open-network-stream name buffer host port))))
-
-(defun imap-open-network-stream (name buffer host &optional port)
-  (open-network-stream name buffer host port))
-
-(defun imap-open-ssl-stream (name buffer host &optional port)
-  (let ((ssl-program-arguments '("-connect" (concat host ":" service)))
-	(proc (open-ssl-stream name buffer host port)))
-    (save-excursion
-      (set-buffer buffer)
-      (goto-char (point-min))
-      (while (not (re-search-forward "^\r*\* OK" nil t))
-	(accept-process-output proc imap-timeout)
-	(goto-char (point-min)))
-      (beginning-of-line)
-      (delete-region (point-min) (point))
-      proc)))
-
-(defun imap-open-imtest-stream (name buffer host &optional port)
-  (let ((process (start-process name (or buffer (current-buffer))
-                                "imtest" "-kp" host 
-				(number-to-string (or port 
-						      imap-default-port)))))
-    (with-current-buffer (process-buffer process)
-      (setq imap-eol "\n")
-      (setq imap-do-login nil) ;; don't login even if kerberos auth fails
-      (when process
-	(message "Opening Kerberized IMAP connection...")
-	;; Result of authentication is a string: __Full privacy protection__
-	(while (and (member (process-status imap-process) '(open run))
-		    (not (or
-			  ;; if everything is ok, this will match...
-			  (re-search-backward "__\\(.*\\)__\n" nil t)
-			  ;; ...errors will match this
-			  (re-search-backward 
-			   "\\. \\([^OA][^KU][^T][^H].*\\)\r\n" nil t))))
-	  (accept-process-output process 1)
-	  (sit-for 1)) ; Yes, this is an oo loop, allow for C-g
-	(let ((response (match-string 1)))
-	  (erase-buffer)
-	  (message "Kerberized IMAP connection: %s" response)
-	  ;; If the __string__ contains "failed" authentication failed
-	  ;; (imtest will bug out if you try to login the usual way, so
-	  ;; close connection with an error)
-	  (when (string-match "failed\\|NO\\|BAD" response)
-	    (mapc 'make-variable-buffer-local imap-locals) ; just in case
-	    ;; XXX logout here (can't use send-command since we don't
-	    ;; have the server opened..)
-	    (delete-process imap-process)
-	    (mapc (lambda (local) (set local (default-value local))) imap-locals)
-	    (erase-buffer)
-	    (error "imtest: %s" response))))
-      process)))
-
 (defun imap-arrival-filter (proc string)
   "Process filter for imap process.  Stow the string, then call the routines
 to try to handle any input.  We need this because we're not guaranteed to
@@ -992,10 +1168,6 @@ VALUE is the value of PROP for each message."
   (with-current-buffer (or buffer (current-buffer))  
    (get (intern (imap-message-to-string id) imap-message-data) prop)))
 
-(defun imap-message-reset (&optional buffer)
-  (with-current-buffer (or buffer (current-buffer)) 
-    (setq imap-message-data (make-vector imap-convenient-folder-prime 0))))
-
 ; Fetches MUST include (UID) otherwise we can't store the results.
 ; NUM is always the logical message number not the UID.  We ignore it.
 (defun imap-cb-fetch (num fetch data)
@@ -1015,15 +1187,6 @@ in `imap-message-data'."
   ;; For some reason found doesn't have parens so we need the &rest
   (imap-folder-set 'search found))
 
-(defun imap-cb-capability (code &rest capabilities)
-  (setq imap-data-capability capabilities))
-
-(defun imap-cb-acl (code group &rest acls)
-  (imap-folder-set 'acl acls (symbol-name group)))
-
-(defun imap-cb-namespace (code &rest namespace)
-  (setq imap-data-namespace namespace))
-
 (defun imap-cb-status (code folder statuses)
 ;  (check-valid-plist statuses)
   (mapc (lambda (c) (imap-folder-set (car c) (cdr c) (symbol-name folder)))
@@ -1060,6 +1223,8 @@ in `imap-message-data'."
 	     (message "IMAP: %s %s" code string))
 	    ((equal var 'UIDNEXT)
 	     (imap-folder-set 'uidnext value))
+	    ((equal var 'UIDNOTSTICKY)
+	     (imap-folder-set 'uidnotsticky value))
             ((equal var 'UNSEEN)
              (imap-folder-set 'unseen value))
             ((equal var 'UIDVALIDITY) ; We should wipe cache clean here only???
