@@ -27,7 +27,7 @@
 ;;;  (setq gnus-secondary-select-methods 
 ;;;      '((nnimap 
 ;;;          "yoyo"
-;;;          (nnimap-server-address "robby.caltech.edu"))))
+;;;          (nnimap-address "robby.caltech.edu"))))
 
 ;;; And inside Gnus, list all available groups with A A (search for 
 ;;; 'nnimap') and subscribe to the mailboxes you are interested in with U. 
@@ -47,6 +47,7 @@
 
 ;;; Todo (roughly in order of priority):
 
+;;;   o What about Gnus's article editing, can we support it?
 ;;;   o Move common IMAP commands to functions in imap.el.
 ;;;   o Verify that we don't use IMAP4rev1 specific things (RFC2060 App B)
 ;;;   o dont uid fetch 1,* in nnimap-retrive-groups (slow)
@@ -95,7 +96,7 @@
 
 (nnoo-declare nnimap) ; we derive from no one
 
-(defconst nnimap-version "nnimap 0.37")
+(defconst nnimap-version "nnimap 0.81")
 
 ;; Various server variables.
 
@@ -137,9 +138,12 @@ delimiter.")
 \"LIST\". LSUB means only retrieve groups marked on the server as
 subscribed.  LIST means every matching group should be retrieved.")
 
-(defvoo nnimap-server-address nil
+(defvoo nnimap-address nil
   "*The name of the IMAP server.  If nil, uses the virtual server's name.")
 
+(defvoo nnimap-server-address nil
+  "Obsolete. Use `nnimap-address'.")
+
 (defvoo nnimap-server-port nil
   "*The port of the IMAP server.  If nil, uses the default port. (143).")
 
@@ -203,20 +207,39 @@ mail belongs in that group.")
 
 ;; Authorization / Privacy variables
 
-;; todo:
-;; better 'smart that open a connection, check the capabilities and
-;; set this variable to the "best" method. easy, but we should fallback
-;; if it doesn't work.
-(defvoo nnimap-auth-method 'smart
+(defvoo nnimap-auth-method nil
+  "Obsolete.")
+
+(defvoo nnimap-stream nil
+  "How nnimap will connect to the server.
+
+The default, nil, will try to use the \"best\" method the server can
+handle.
+
+Change this if
+
+1) you want to connect with SSL. The SSL integration with IMAP is
+   brain-dead so you'll have to tell it specifically.
+
+2) your server is more capable than your environment -- i.e. your
+   server accept Kerberos login's but you haven't installed the
+   `imtest' program or your machine isn't configured for Kerberos.
+
+Possible choices: kerberos4, ssl, network")
+
+(defvoo nnimap-authenticator nil
   "How nnimap authenticate itself to the server.
 
-The default, `smart', will try to use the \"best\" method the server
-can handle.
+The default, nil, will try to use the \"best\" method the server can
+handle.
 
-Unfortunely, the `smart' option isn't all that clever and will not
-recognize KERBEROS_V4 or SSL so you should set this if you use them.
+There is only one reason for fiddling with this variable, and that is
+if your server is more capable than your environment -- i.e. you
+connect to a server that accept Kerberos login's but you haven't
+installed the `imtest' program or your machine isn't configured for
+Kerberos.
 
-Available options: smart, login, cram-md5, kerberos4 and ssl.")
+Possible choices: kerberos4, cram-md5, login, anonymous.")
 
 (defcustom nnimap-authinfo-file "~/.authinfo"
   "Authorization information for IMAP servers. In .netrc format."
@@ -269,7 +292,6 @@ Available options: smart, login, cram-md5, kerberos4 and ssl.")
           nnimap-close-group
           nnimap-close-server
           nnimap-open-server  
-	  nnimap-authenticate
 	  nnimap-expunge-close-group
 	  nnimap-date-days-ago
 	  nnimap-time-substract
@@ -351,10 +373,8 @@ BODY."
 (deffoo nnimap-retrieve-headers (articles &optional group server fetch-old)
   (when (nnimap-possibly-change-group group server)
     (with-current-buffer nntp-server-buffer (erase-buffer))
-    (let* (compressed (uncompressed (gnus-uncompress-sequence articles)))
-      (when (numberp fetch-old)
-        (setcar uncompressed (- (car uncompressed) fetch-old)))
-      (setq compressed (gnus-compress-sequence uncompressed t))
+    (let* ((uncompressed (gnus-uncompress-sequence articles))
+          (compressed (gnus-compress-sequence uncompressed t)))
       (with-current-buffer nnimap-server-buffer
 	;; Reset message info, this makes sure we return 423 on
 	;; articles that were removed by another client while we
@@ -384,36 +404,6 @@ BODY."
 	      uncompressed)
 	'headers))))
 
-(defun nnimap-authenticate (server &optional buffer)
-  (with-current-buffer (or buffer (current-buffer))
-    ;; Login
-    (let (list alist ok)
-      (and (fboundp 'gnus-parse-netrc)
-	   (setq list (gnus-parse-netrc nnimap-authinfo-file)
-		 alist (or (and (gnus-netrc-get (gnus-netrc-machine 
-						 list server) "machine")
-				(gnus-netrc-machine list server))
-			   (gnus-netrc-machine list nnimap-server-address))
-		 imap-username (gnus-netrc-get alist "login")
-		 imap-password (gnus-netrc-get alist "password")))
-      (cond ((eq nnimap-auth-method 'smart)
-	     (and (memq 'AUTH=CRAM-MD5 (imap-capability-get))
-		  (setq imap-auth-method 'imap-authenticate-cram-md5)))
-	    ((eq nnimap-auth-method 'md5) 
-	     (setq imap-auth-method 'imap-authenticate-cram-md5)))
-      (setq ok (imap-authenticate server buffer))
-      ;; only loop if 1) the server didn't validate us AND 2) the user
-      ;; actually typed something
-      (while (and (not ok)
-		  imap-username
-		  (not (not imap-password))) ;; no password in backtraces
-	(message "Bad Password for %s, Try again." server)
-	(sleep-for 2)
-	(setq ok (imap-authenticate server buffer)))
-      ok)))
-
-;; todo:
-;;   only condition-case on imap-open-stream??
 (deffoo nnimap-open-server (server &optional defs)
   (or (and (nnimap-server-opened server)
 	   (nnoo-change-server 'nnimap server defs)
@@ -421,60 +411,47 @@ BODY."
                (error "this shouldn't happen")
              t))
       (progn
-        (unless (assq 'nnimap-server-address defs)
-          (push (list 'nnimap-server-address server) defs))
+        (unless (assq 'nnimap-address defs)
+	  (if (assq 'nnimap-server-address defs)
+	      (push (list 'nnimap-address (cadr (assq 'nnimap-server-address
+						      defs))) defs)
+	    (push (list 'nnimap-address server) defs)))
         (unless (assq 'nnimap-server-buffer defs)
-          (push (list 'nnimap-server-buffer (concat " *nnimap* " server)) defs))
+          (push (list 'nnimap-server-buffer (concat " *nnimap* " server))
+		defs))
 	(nnoo-change-server 'nnimap server defs)
-	(condition-case err
-	    ;; todo: do clever things here if nnimap-auth-method is 'smart
-	    (let ((imap-open-stream (cond ((eq nnimap-auth-method 
-					       'kerberos4)
-					   'imap-open-imtest-stream)
-					  ((eq nnimap-auth-method
-					       'ssl)
-					   'imap-open-ssl-stream))))
-	      (when (imap-open-server nnimap-server-address 
-				      nnimap-server-port
-				      nnimap-server-buffer 
-				      nnimap-imap-defs)
-		;; we only support imap4.*
-		(unless (or (memq 'IMAP4 (imap-capability-get
-					  nnimap-server-buffer))
-			    (memq 'IMAP4rev1 (imap-capability-get
-					      nnimap-server-buffer))
-			    (memq 'IMAP4REV1 (imap-capability-get
-					      nnimap-server-buffer)))
-		  (imap-close-server nnimap-server-buffer)
-		  (nnheader-report 
-		   'nnimap 
-		   "Sorry, %s is not a IMAP4(rev1) server." 
-		   nnimap-server-address)
-		  (sit-for 2)
-		  ;; Make sure we close the server
-		  (signal 'quit))
-		;; authenticate ourself
-		(when (nnimap-authenticate server nnimap-server-buffer)
-		  (push (cons server nnimap-server-buffer)
-			nnimap-server-buffer-alist))))
-	  ;; Assume error comes from host unknow so we are not connected yet
-	  (error (kill-buffer nnimap-server-buffer)
-		 (nnheader-report 'nnimap (cdr err)))
-	  (quit (let (imap-last-status)
-		  (imap-close-server nnimap-server-buffer)
-		  (unless (get-buffer nnimap-server-buffer)
-		    (kill-buffer nnimap-server-buffer))
-		  (nnheader-report 'nnimap (cdr err))))))))
+	(if (not (imap-open nnimap-address nnimap-server-port
+			    nnimap-stream nnimap-authenticator 
+			    nnimap-server-buffer))
+	    (nnheader-report 'nnimap "Could not connect to server %s" server)
+	  (unless (or (imap-capability 'IMAP4 nnimap-server-buffer)
+		      (imap-capability 'IMAP4rev1 nnimap-server-buffer)
+		      (imap-capability 'IMAP4REV1 nnimap-server-buffer))
+	    (imap-close nnimap-server-buffer)
+	    (nnheader-report 'nnimap "Server %s not IMAP4" server))
+	  (let (list alist user passwd)
+	    (and (fboundp 'gnus-parse-netrc)
+		 (setq list (gnus-parse-netrc nnimap-authinfo-file)
+		       alist (or (and (gnus-netrc-get (gnus-netrc-machine 
+						       list server) "machine")
+				      (gnus-netrc-machine list server))
+				 (gnus-netrc-machine list nnimap-address))
+		       user (gnus-netrc-get alist "login")
+		       passwd (gnus-netrc-get alist "password")))
+	    (if (imap-authenticate nnimap-server-buffer user passwd)
+		(push (cons server nnimap-server-buffer)
+		      nnimap-server-buffer-alist)
+	      (imap-close nnimap-server-buffer)
+	      (nnheader-report 'nnimap "Could not authenticate to %s"
+			       server)))))))
 
 (deffoo nnimap-close-server (&optional server)
-  (let ((s-b (assoc 
-              (setq server (or server (nnoo-current-server 'nnimap)))
-              nnimap-server-buffer-alist)))
+  (let ((s-b (assoc (setq server (or server (nnoo-current-server 'nnimap)))
+		    nnimap-server-buffer-alist)))
     (when s-b
       (setq nnimap-server-buffer nil)
       (setq nnimap-server-buffer-alist (delq s-b nnimap-server-buffer-alist))
-      (imap-close-server (cdr s-b))
-      (kill-buffer (cdr s-b)))
+      (imap-close (cdr s-b)))
     (nnoo-close-server 'nnimap server)))
 
 (deffoo nnimap-request-close ()
@@ -488,15 +465,12 @@ BODY."
                       nnimap-server-buffer-alist))))
     (if (and (gnus-buffer-live-p buffer)
 	     (gnus-buffer-live-p nntp-server-buffer))
-	(let ((running (imap-server-opened buffer)))
+	(let ((running (imap-opened buffer)))
 	  ;; clean up autologouts
 	  (unless running
 	    (nnimap-close-server server))
 	  running))))
 
-;  (and (nnoo-server-opened 'nnimap server)
-;       (imap-server-opened nnimap-server-buffer)))
-
 (deffoo nnimap-status-message (&optional server)
   (let ((buffer (cdr (assoc 
                       (setq server (or server (nnoo-current-server 'nnimap)))
@@ -618,7 +592,7 @@ BODY."
       (gnus-message 7 "Generating active list for nnimap group %s" group)
       (cond 
        ((eq nnimap-group-list-speed 'slow)
-	(when (imap-select-folder group)
+	(when (imap-mailbox-select group)
 	  (let ((exists (imap-folder-get 'EXISTS))
 		articles)
 	    (if (eq 0 exists)
@@ -665,8 +639,9 @@ BODY."
 
 (defun nnimap-pattern-to-list-arguments (pattern)
   (mapcar (lambda (p) (if (consp p)
-			  p 
-			(cons "\"\"" p)))
+			  (cons (concat "\"" (car p) "\"")
+				(concat "\"" (cdr p) "\""))
+			(cons "\"\"" (concat "\"" p "\""))))
 	  (if (and (listp pattern)
 		   (listp (cdr pattern)))
 	      pattern
@@ -686,7 +661,7 @@ BODY."
                                 (car pattern) " "
                                 (cdr pattern))))))
       (let ((nnimap-group-list-speed 'fast))
-	(imap-folder-map 'nnimap-request-list-mapper))
+	(imap-mailbox-map 'nnimap-request-list-mapper))
       t)))
   
 
@@ -717,11 +692,12 @@ BODY."
 ;; we can't have Gnus call it all the time. Instead, it's called by
 ;; nnimap-request-group.
 (deffoo nnimap-request-update-info-internal (group info &optional server)
-  (when (nnimap-possibly-change-group group server) ;; SELECT
-    ;; We reset the uidvalidity here because we are about to do a full resync.
+  ;; We reset the uidvalidity here because we are about to do a full resync.
+  (when info ;; nil info when group doesn't exist or has just been created.
+             ;; Gnus will create it for us later.
     (gnus-info-set-params info (nnimap-update-alist-soft
-				'uidvalidity nil (gnus-info-params info)) t)
-    
+                               'uidvalidity nil (gnus-info-params info)) t)
+  (when (nnimap-possibly-change-group group server) ;; SELECT
     (with-current-buffer nnimap-server-buffer
       (gnus-message 5 "Updating info for mailbox %s" group)
       
@@ -768,7 +744,7 @@ BODY."
 		   (imap-search (nnimap-mark-to-predicate (cdr pred))))
 		  (gnus-info-marks info))
 		 t)))
-	      gnus-article-mark-lists)))
+	      gnus-article-mark-lists))))
   info)
 
 ;; Respond to articles with mail
@@ -780,7 +756,7 @@ BODY."
   (when (nnimap-ok-p (nnimap-send-command-wait
 		      (format "UID COPY %d %s" article to-group)))
     (setq nnimap-need-expunge t)
-    (if (imap-store-flags-add (format "%d" article) "\\Seen \\Deleted")
+    (if (imap-message-flags-add (format "%d" article) "\\Seen \\Deleted")
 	(message "IMAP split moved %s:%s:%d to %s" server group
 		 article to-group)
       (error "IMAP flag store failed: you may have unread mail marked as read!"))))
@@ -846,7 +822,7 @@ BODY."
 		      ;; move to first matching box, if any
 		      (nnimap-split-move-article article inbox 
 						 (car groups) server)))))))
-	  (when (imap-select-folder inbox) ;; just in case
+	  (when (imap-mailbox-select inbox) ;; just in case
 	    ;; todo: UID EXPUNGE (if available) to remove splitted articles
 	    (nnimap-expunge-close-group)))
 	t))))
@@ -907,13 +883,13 @@ BODY."
 		(setq pred (cons 'tick pred)))
 	    (when (and range pred)
 	      (cond ((eq what 'del)
-		     (imap-store-flags-del (nnimap-range-to-string range)
+		     (imap-message-flags-del (nnimap-range-to-string range)
 					   (nnimap-mark-to-flag pred nil t)))
 		    ((eq what 'add)
-		     (imap-store-flags-add (nnimap-range-to-string range)
+		     (imap-message-flags-add (nnimap-range-to-string range)
 					   (nnimap-mark-to-flag pred nil t)))
 		    ((eq what 'set)
-		     (imap-store-flags-set (nnimap-range-to-string range)
+		     (imap-message-flags-set (nnimap-range-to-string range)
 					   (nnimap-mark-to-flag pred nil t)))))))
 	(gnus-message 7 "Setting marks in %s:%s...done" 
 		      (nnoo-current-server 'nnimap) group)))))
@@ -954,7 +930,7 @@ BODY."
       (with-current-buffer nnimap-server-buffer
 	(if force
 	    ;; add delete flag to article
-	    (when (imap-store-flags-add (nnimap-range-to-string artseq) 
+	    (when (imap-message-flags-add (nnimap-range-to-string artseq) 
 					"\\Deleted")
 	      (setq nnimap-need-expunge t)
 	      (setq articles nil))
@@ -963,8 +939,8 @@ BODY."
 			  nnmail-expiry-wait)))
 	    (cond ((eq days 'immediate)
 		   ;; add delete flag to article
-		   (when (imap-store-flags-add (nnimap-range-to-string artseq) 
-					       "\\Deleted")
+		   (when (imap-message-flags-add 
+			  (nnimap-range-to-string artseq) "\\Deleted")
 		     (setq nnimap-need-expunge t)
 		     (setq articles nil)))
 		  ((numberp days)
@@ -979,7 +955,7 @@ BODY."
 					  (nnimap-date-days-ago days))))
 		   (let ((imap-cb-fetch-hook 
 			  'nnimap-request-expire-articles-progress))
-		     (when (and oldarts (imap-store-flags-add 
+		     (when (and oldarts (imap-message-flags-add 
 					 (nnimap-range-to-string 
 					  (gnus-compress-sequence oldarts))
 					 "\\Deleted"))
@@ -1009,7 +985,7 @@ BODY."
 (deffoo nnimap-request-accept-article (group &optional server last)
   (when (nnimap-possibly-change-server server)
     ;; We assume article is appended as UIDNEXT if no UIDPLUS support.
-    (when (or (memq 'UIDPLUS (imap-capability-get nnimap-server-buffer))
+    (when (or (imap-capability 'UIDPLUS nnimap-server-buffer)
 	      (nnimap-ok-p (nnimap-send-command-wait
 			    (concat "STATUS " group " (UIDNEXT)")
 			    nnimap-server-buffer)))
@@ -1023,8 +999,7 @@ BODY."
 			  (list (concat "APPEND " group " ") 
 				(current-buffer))
 			  nnimap-server-buffer))
-	(let ((high (if (memq 'UIDPLUS (imap-capability-get
-					nnimap-server-buffer))
+	(let ((high (if (imap-capability 'UIDPLUS nnimap-server-buffer)
 			(cdr (imap-folder-get 'appenduid nil 
 					      nnimap-server-buffer))
 		      (imap-folder-get 'UIDNEXT group
@@ -1172,16 +1147,16 @@ on articles, in GROUP."
 	     (when nnimap-need-expunge 
 	       (setq nnimap-need-expunge nil)
 	       (imap-send-command "EXPUNGE"))
-	     (imap-expunge-close-folder))
+	     (imap-mailbox-close))
 	    ((eq nnimap-expunge-on-close 'never)
-	     (imap-unselect-folder))
+	     (imap-mailbox-unselect))
 	    ((eq nnimap-expunge-on-close 'ask)
 	     (when (imap-search "DELETED")
 	       (if (gnus-y-or-n-p (format "Expunge articles in group `%s'? "
 					  imap-current-folder))
 		   (and (nnimap-ok-p (nnimap-send-command-wait "EXPUNGE"))
-			(imap-expunge-close-folder))
-		 (imap-unselect-folder)))))))
+			(imap-mailbox-close))
+		 (imap-mailbox-unselect)))))))
   (not imap-current-folder))
 
 (defun nnimap-possibly-change-server (server)
@@ -1200,7 +1175,7 @@ nil, change to current server."
 	  (if (and imap-current-folder
 		   (not (string= group imap-current-folder)))
 	      (nnimap-expunge-close-group))
-	  (when (imap-select-folder group nil 
+	  (when (imap-mailbox-select group nil 
 				    (gnus-ephemeral-group-p groupname))
 	    ;; check/set UIDVALIDITY
 	    (let ((new-uid (imap-folder-get 'uidvalidity))
@@ -1216,7 +1191,7 @@ nil, change to current server."
 						new-uid)
 		    (message "UIDVALIDITY clash. Old value `%s', new `%s'"
 			     old-uid new-uid)
-		    (imap-unselect-folder))))))))
+		    (imap-mailbox-unselect))))))))
       imap-current-folder)))
 
 ;;; Gnus functions
@@ -1247,7 +1222,7 @@ nil, change to current server."
     (unless (eq 'nnimap (car (setq method (gnus-find-method-for-group group))))
       (error "ACL editing only available for nnimap groups"))
     (when (nnimap-possibly-change-server (cadr method))
-      (unless (memq 'ACL (imap-capability-get nnimap-server-buffer))
+      (unless (imap-capability 'ACL nnimap-server-buffer)
 	(error "Your server does not support ACL editing"))
       (gnus-edit-form (with-current-buffer nnimap-server-buffer
 			(imap-folder-set 'acl nil mailbox)
